@@ -16,6 +16,7 @@
 
 import stat
 import platform, subprocess, operator, os, shutil, re
+import collections
 
 from glob import glob
 
@@ -121,16 +122,18 @@ class File:
         self.is_built = is_built
         self.subdir = subdir
         self.fname = fname
+        assert(isinstance(self.subdir, str))
+        assert(isinstance(self.fname, str))
 
     def __str__(self):
-        return os.path.join(self.subdir, self.fname)
+        return self.relative_name()
 
     def __repr__(self):
         ret = '<File: {0}'
         if not self.is_built:
             ret += ' (not built)'
         ret += '>'
-        return ret.format(os.path.join(self.subdir, self.fname))
+        return ret.format(self.relative_name())
 
     @staticmethod
     def from_source_file(source_root, subdir, fname):
@@ -148,15 +151,15 @@ class File:
 
     def rel_to_builddir(self, build_to_src):
         if self.is_built:
-            return os.path.join(self.subdir, self.fname)
+            return self.relative_name()
         else:
             return os.path.join(build_to_src, self.subdir, self.fname)
 
     def absolute_path(self, srcdir, builddir):
+        absdir = srcdir
         if self.is_built:
-            return os.path.join(builddir, self.subdir, self.fname)
-        else:
-            return os.path.join(srcdir, self.subdir, self.fname)
+            absdir = builddir
+        return os.path.join(absdir, self.relative_name())
 
     def endswith(self, ending):
         return self.fname.endswith(ending)
@@ -173,6 +176,15 @@ class File:
     def relative_name(self):
         return os.path.join(self.subdir, self.fname)
 
+def get_meson_script(env, script):
+    '''
+    Given the path of `meson.py`/`meson`, get the path of a meson script such
+    as `mesonintrospect` or `mesontest`.
+    '''
+    meson_py = env.get_build_command()
+    (base, ext) = os.path.splitext(meson_py)
+    return os.path.join(os.path.dirname(base), script + ext)
+
 def get_compiler_for_source(compilers, src):
     for comp in compilers:
         if comp.can_compile(src):
@@ -210,6 +222,10 @@ def is_windows():
     platname = platform.system().lower()
     return platname == 'windows' or 'mingw' in platname
 
+def is_cygwin():
+    platname = platform.system().lower()
+    return platname.startswith('cygwin')
+
 def is_debianlike():
     return os.path.isfile('/etc/debian_version')
 
@@ -452,25 +468,22 @@ def replace_if_different(dst, dst_tmp):
     else:
         os.unlink(dst_tmp)
 
-def stringintlistify(item):
-    if isinstance(item, (str, int)):
+def typeslistify(item, types):
+    '''
+    Ensure that type(@item) is one of @types or a
+    list of items all of which are of type @types
+    '''
+    if isinstance(item, types):
         item = [item]
     if not isinstance(item, list):
-        raise MesonException('Item must be a list, a string, or an int')
+        raise MesonException('Item must be a list or one of {!r}'.format(types))
     for i in item:
-        if not isinstance(i, (str, int, type(None))):
-            raise MesonException('List item must be a string or an int')
+        if i is not None and not isinstance(i, types):
+            raise MesonException('List item must be one of {!r}'.format(types))
     return item
 
 def stringlistify(item):
-    if isinstance(item, str):
-        item = [item]
-    if not isinstance(item, list):
-        raise MesonException('Item is not a list')
-    for i in item:
-        if not isinstance(i, str):
-            raise MesonException('List item not a string.')
-    return item
+    return typeslistify(item, str)
 
 def expand_arguments(args):
     expended_args = []
@@ -492,6 +505,7 @@ def expand_arguments(args):
 
 def Popen_safe(args, write=None, stderr=subprocess.PIPE, **kwargs):
     p = subprocess.Popen(args, universal_newlines=True,
+                         close_fds=False,
                          stdout=subprocess.PIPE,
                          stderr=stderr, **kwargs)
     o, e = p.communicate(write)
@@ -672,3 +686,18 @@ def get_filenames_templates_dict(inputs, outputs):
         if values['@OUTDIR@'] == '':
             values['@OUTDIR@'] = '.'
     return values
+
+class OrderedSet(collections.OrderedDict):
+    '''
+    A 'set' equivalent that preserves the order in which items are added.
+
+    This is a hack implementation that wraps OrderedDict. It may not be the
+    most efficient solution and might need fixing to override more methods.
+    '''
+    def __init__(self, iterable=None):
+        if iterable:
+            self.update(iterable)
+
+    def update(self, iterable):
+        for item in iterable:
+            self[item] = True
